<?php

namespace Goldenscarab\Modulus\Service\Template\Compose;

use Illuminate\Support\Facades\Request;
use \Illuminate\Support\HtmlString;
use Illuminate\Support\FacadesRequest;


class Template3Bt4
{
    private $route_prefix;

    private $route_param;

    private $table_id;

    private $table_class;

    private $can_prefix;

    private $source;

    private $_menu;

    private $_rows;

    private $_columns;


    public function __construct($attrs = [])
    {
        $this->setAttrs($attrs);
    }

    /**
     * Retourne le rendu d'un tableau liste
     * @return string le rendu du tableau liste
     */
    public function renderList(): string
    {
        $template  = '<div class="table-responsive-lg" style="min-height: 250px;">' . PHP_EOL;
        $template .= '    <table id="%s" class="%s">' . PHP_EOL;
        $template .= '        <thead>' . PHP_EOL;
        $template .= '            <tr>' . PHP_EOL;
        $template .=                '%s';
        $template .= '            </tr>' . PHP_EOL;
        $template .= '        </thead>' . PHP_EOL;
        $template .= '        <tbody>' . PHP_EOL;
        $template .=             '%s';
        $template .= '        </tbody>' . PHP_EOL;
        $template .= '    </table>';
        $template .= '</div>';


        $render = vsprintf($template, [
            empty($this->table_id) ? 'table-' . str_slug($this->route_prefix) : $this->table_id,
            trim('table table-sm table-hover table-striped ' . $this->table_class),
            $this->renderIndent($this->renderListHead(), 3),
            $this->renderIndent($this->renderListBody(), 2),
        ]);

        return $render;
    }

    /**
     * Retourne le rendu des entêtes de colonnes d'un tableau liste
     * @return string Le rendu des entêtes de colonnes
     */
    public function renderListHead(): string
    {
        $template = '<th scope="col" class="text-nowrap">%s%s</th>' . PHP_EOL;

        $render = '';

        $columns = $this->_columns;

        // Colonne titre
        if (is_iterable($columns)) {

            foreach ($columns as $column) {

                $sortable = data_get($column, 'sortable');

                $render .= vsprintf($template, [
                    data_get($column, 'label'),
                    is_null($sortable) ? '' : col_sort($sortable)
                ]);
            }
        }

        // Colonne menu
        $render .= vsprintf($template, ['','', '']);

        return $render;
    }

    /**
     * Retourne le rendu des lignes d'un tableau liste
     * @return string Le rendu des lignes du tableau
     */
    public function renderListBody(): string
    {
        $template_row  = '<tr scope="row"%s>' . PHP_EOL;
        $template_row .=     '%s';
        $template_row .= '</tr>' . PHP_EOL;
        $template_cel  = '<td%s>%s</td>' . PHP_EOL;

        $render     = '';
        $collection = $this->source;
        $columns    = $this->_columns;
        $rows       = $this->_rows;

        if (is_iterable($collection)) {
            if (count($collection) == 0) {
                $render_cels_row = vsprintf($template_cel, [
                    ' colspan="' . (count($columns) + 1) .'"',
                    'Aucun élément disponible'
                ]);

                $render .= vsprintf($template_row, [
                    null,
                    $this->renderIndent($render_cels_row, 1)
                ]);
            } else {

                foreach ($collection as $item) {
                    $render_cels_row = '';
                    if (is_iterable($columns)) {
                        foreach ($columns as $column) {

                            $attrs      = collect($column)->except(['label', 'row', 'value', 'sortable', 'id']);
                            $attributes = $this->renderAttributes($attrs, $item);
                            $value      = $this->getValue(data_get($column, 'value'), $item);

                            // Création de la cellule
                            $render_cels_row .= vsprintf($template_cel, [
                                $attributes,
                                $value
                            ]);
                        }

                        // Ajout du menu en fin de ligne
                        $render_cels_row .= vsprintf($template_cel, [
                            null,
                            data_get($this->_menu, 'display') ? $this->renderListMenuItems($item) : ''
                        ]);

                    }
                    $render .= vsprintf($template_row, [
                        $this->renderAttributes($rows, $item),
                        $this->renderIndent($render_cels_row, 1)
                    ]);
                }
            }
        }

        return $render;
    }

    /**
     * Retourne le rendu du menu item
     * @param  Object $source Source des données
     * @return string         Le rendu
     */
    private function renderListMenuItems($source): string
    {
        $template  = '<div class="dropdown">' . PHP_EOL;
        $template .= '    <button class="btn btn-default btn-xs dropdown-toggle no-caret square-to-circle item-option" type="button" data-toggle="dropdown" aria-haspopup="true" aria-expanded="false">' . PHP_EOL;
        $template .= '        <i class="fa fa-cog" aria-hidden="true"></i>' . PHP_EOL;
        $template .= '    </button>' . PHP_EOL;
        $template .= '    <div class="dropdown-menu dropdown-menu-right">' . PHP_EOL;
        $template .= '%s';
        $template .= '%s';
        $template .= '%s';
        $template .= '%s';
        $template .= '%s';
        $template .= '%s';
        $template .= '    </div>' . PHP_EOL;

        $render = '';

        $render = vsprintf($template, [
            $this->renderIndent($this->renderListMenuItemRead($source), 2),
            $this->renderIndent($this->renderListMenuItemUpdate($source), 2),
            $this->renderIndent($this->renderListMenuItemTag($source), 2),
            $this->renderIndent($this->renderListMenuItemMore($source), 2),
            $this->renderIndent($this->renderListMenuItemDuplicate($source), 2),
            $this->renderIndent($this->renderListMenuItemDelete($source), 2),
        ]);

        return $render;
    }

    /**
     * Retourne l'element voir du menu item
     * @param  Object $source Source des données
     * @return string         Le rendu
     */
    private function renderListMenuItemRead($source): string
    {
        $render = '';

        if (auth()->user()->can($this->can_prefix . '-read')) {

            if (data_get($this->_menu, 'is_readable')) {
                $template  = '<a class="dropdown-item" href="%s">' . PHP_EOL;
                $template .= '    <i class="fa fa-eye mr-1"></i>Voir' . PHP_EOL;
                $template .= '</a>' . PHP_EOL;

                if (empty($this->route_param)) {
                    $render = sprintf($template, route($this->route_prefix . '.' . 'read', array_merge([$source->id], Request::query())));
                } else {
                    $render = sprintf($template, route($this->route_prefix . '.' . 'read', array_merge([$this->route_param, $source->id ], Request::query())));
                }
            }
        }

        return $render;
    }

    /**
     * Retourne l'element modifier du menu item
     * @param  Object $source Source des données
     * @return string         Le rendu
     */
    private function renderListMenuItemUpdate($source): string
    {
        $render = '';

        if (auth()->user()->can($this->can_prefix . '-update')) {

            if (data_get($this->_menu, 'is_editable')) {
                $template  = '<a class="dropdown-item" href="%s">' . PHP_EOL;
                $template .= '    <i class="fa fa-pencil-square-o mr-1"></i>Modifier' . PHP_EOL;
                $template .= '</a>' . PHP_EOL;

                if (empty($this->route_param)) {
                    $render = sprintf($template, route($this->route_prefix . '.' . 'update', array_merge([$source->id], Request::query())));
                } else {
                    $render = sprintf($template, route($this->route_prefix . '.' . 'update', array_merge([$this->route_param, $source->id ], Request::query())));
                }
            }
        }

        return $render;
    }

    /**
     * Retourne l'element tag du menu item
     * @param  Object $source Source des données
     * @return string         Le rendu
     */
    private function renderListMenuItemTag($source): string
    {
        $render = '';

        if (auth()->user()->hasRole('admin')) {

            if (count(data_get($this->_menu, 'tags')) > 0) {
                $template  = '<a class="dropdown-item clipboard-copy" href="#" title="Copier le Tag" data-clipboard-action="copy" data-clipboard-text="%s">' . PHP_EOL;
                $template .= '    <i class="fa fa-clipboard mr-1"></i>%s' . PHP_EOL;
                $template .= '</a>' . PHP_EOL;

                foreach (data_get($this->_menu, 'tags') as $tag) {
                    $tag_value = $this->getValue($tag, $source);
                    $render .= vsprintf($template, [$tag_value, $tag_value]);
                }
            }
        }

        return $render;
    }

    /**
     * Retourne les éléments supplémentaires du menu item
     * @param  Object $source Source des données
     * @return string         Le rendu
     */
    private function renderListMenuItemMore($source): string
    {
        $render = '';

        if (auth()->user()->can($this->can_prefix . '-read')) {

            $menu_more = data_get($this->_menu, 'more');

            if (count($menu_more) > 0) {
                $template  = '<a class="dropdown-item%s"%s>' . PHP_EOL;
                $template .= '    %s' . PHP_EOL;
                $template .= '</a>' . PHP_EOL;

                foreach ($menu_more as $menu_item) {

                    $href     = data_get($menu_item, 'href');
                    $name     = data_get($menu_item, 'name');
                    $class    = data_get($menu_item, 'class');

                    $attrs = collect($menu_item)->except(['label', 'name', 'class']);

                    // Rendu de l'élément
                    $render .= vsprintf($template, [
                        empty($class) ? '' :  ' ' . $class,
                        $this->renderAttributes($attrs, $source),
                        $menu_item['label']
                    ]);

                }
            }
        }

        return $render;
    }

    /**
     * Retourne l'element dupliqué du menu item
     * @param  Object $source Source des données
     * @return string         Le rendu
     */
    private function renderListMenuItemDuplicate($source): string
    {
        $render = '';

        if (auth()->user()->can($this->can_prefix . '-create')) {

            if (data_get($this->_menu, 'is_duplicable')) {
                $template = '<a class="dropdown-item" href="%s">' . PHP_EOL;
                $template .= '    <i class="fa fa-clone mr-1"></i>Dupliquer' . PHP_EOL;
                $template .= '</a>' . PHP_EOL;

                if (empty($this->route_param)) {
                    $render = sprintf($template, route($this->route_prefix . '.' . 'duplicate', array_merge([$source->id], Request::query())));
                } else {
                    $render = sprintf($template, route($this->route_prefix . '.' . 'duplicate', array_merge([$this->route_param, $source->id], Request::query())));
                }
            }

        }

        return $render;
    }

    /**
     * Retourne l'element supprimer du menu item
     * @param  Object $source Source des données
     * @return string         Le rendu
     */
    private function renderListMenuItemDelete($source): string
    {
        $render = '';

        if (auth()->user()->can($this->can_prefix . '-delete')) {

            if (data_get($this->_menu, 'is_deletable')) {
                $template  = '<div class="dropdown-divider"></div>' . PHP_EOL;
                $template .= '<a class="dropdown-item delete" href="#" data-href="%s">' . PHP_EOL;
                $template .= '    <i class="fa fa-trash mr-1"></i>Supprimer' . PHP_EOL;
                $template .= '</a>' . PHP_EOL;

                if (empty($this->route_param)) {
                    $render = sprintf($template, route($this->route_prefix . '.' . 'delete', array_merge([$source->id], Request::query())));
                } else {
                    $render = sprintf($template, route($this->route_prefix . '.' . 'delete', array_merge([$this->route_param, $source->id], Request::query())));
                }
            }

        }

        return $render;
    }


    /**
     * Retourne le rendu d'attributs avec leur valeur
     * @param  array  $attrs Les attributs avec la colonne de l'object conteant la valeur
     * @param  Object $item  Object contenant la valeur de l'attribut
     * @return string        Le rendu des attributs avec leur valeur
     */
    private function renderAttributes($attrs = [], $source = null): string
    {
        $render = '';

        if (is_iterable($attrs)) {
            $template = ' %s="%s"';
            foreach ($attrs as $name => $data) {

                if (is_null($source) || !is_array($data)) {
                    $value = $data;
                } else if (is_array($data)) {
                    $value = $this->getValue($data, $source);
                }

                $render .= vsprintf($template, [$name, $value]);

            }
        }

        return $render;
    }

    /**
     * Vérifier la validité d'un paramètre condition
     * @param  array|null $condition La condition si elle est disponible
     * @param  Objet $datas     Les données dans lesquelles aller chercher la valeur à controler
     * @return bollean          True si la condition est valide, sinon false
     */
    private function checkCondition($condition, $datas): string
    {
        $data      = mixte_get($datas, mixte_get($condition, 'key'));
        $value     = mixte_get($condition, 'value');
        $operator  = mixte_get($condition, 'operator');
        $default   = true;

        switch ($operator) {
            case '==' :
                $validated = $data == $value ? true : false;
                break;
            case '>' :
                $validated = $data > $value ? true : false;
                break;
            case '<' :
                $validated = $data < $value ? true : false;
                break;
            case '!=' :
                $validated = $data != $value ? true : false;
                break;
            case '<=' :
                $validated = $data <= $value ? true : false;
                break;
            case '>=' :
                $validated = $data >= $value ? true : false;
                break;
            default :
                $validated = $default;
        }

        return $validated;
    }

    /**
     * Récupération d'une valeur en fonction de sa composition
     * @param  Mixed $value             array pour une valeur dynamique ou string|int|float|bool pour une valeur statique
     * @param  Object $datas            Objet dans lequel aller chercher la valeur si dynamique
     * @param  boolean $strip           Supprimer les balises HTML dela valeur résultante
     * @return Mixed                    La valeur
     */
    private function getValue($value, $datas = null, $strip = false)
    {
        $_value = null;

        if (is_array($value)) {

            // Si le tableau n'est pas de type associatif ['my-key'] devient ['key' => 'my-key']
            if(array_keys($value) === range(0, count($value) - 1) && count($value) == 1) {
                $value = ['key' => $value[0]];
            }

            $condition = data_get($value, 'condition');

            if ($this->checkCondition($condition, $datas)) {

                $_value = $this->dynamicFormatedValue(
                    data_get($value, 'template'),
                    data_get($value, 'key'),
                    data_get($value, 'value'),
                    data_get($value, 'callback'),
                    $datas,
                    $strip
                );
            }
        } else {
            $_value = $value;
        }

        return $_value;
    }

    /**
     * Retourne le rendu formaté d'une valeur dynamique
     * @param  string $template         Le template servant au formatage
     * @param  string|array $keys       Le ou les nom de colonnes
     * @param  Mix $default             Valeur par défaut
     * @param  string $callback         Fonction callback à appeler pour traiter la valeur
     * @param  Mix $source              Objet source pour récupérer la valeur dynamique
     * @param  boolean $strip           Supprimer les balises HTML
     * @return string                   Le rendu
     */
    private function dynamicFormatedValue($template = null, $keys = null, $default = null, $callback = null, $source, $strip = false)
    {
        // Création du texte à afficher
        $datas = [];

        // Convertion en tableau si besoin
        if (!is_iterable($keys)) {
           $keys = [$keys];
        }

        // Traitement du tableau de valeur à afficher
        foreach ($keys as $key => $column) {

            if (is_null($column)) {
                $data = $default;
            } else {
                $data = mixte_get($source, $column, $default);
            }

            // Callback array
            if (is_array($callback) && isset($callback[$key]) && !is_null($callback[$key])) {
                $data = $callback[$key]($data);
            }

            $datas[] = $strip ? strip_tags($data) : $data;
        }

        // Traitement des % seuls pour éviter les conflits
        if (!empty($template)) {
            $template = preg_replace('/(%)[^0-9sd.\']/i', '%%"', $template);
        }

        // Traitement template & callback
        if (is_string($callback)) {
            if (empty($template)) {
                $string = call_user_func_array($callback, $datas);
            } else {
                $string = vsprintf($template, $datas);
                $string = $callback($string);
            }

        } else if (!empty($template)) {
            $string = vsprintf($template, $datas);
        } else {
            $string = $datas[0];
        }

        return $string;
    }

    /**
     * Retourne le rendu d'une chaine de caractère en HTML
     * @param  String $string La chains de caractères à traiter
     * @return String         Le rendu HTML
     */
    public function renderHTML($string): string
    {
        return new HtmlString($string);
    }

    /**
     * Retourne le rendu avec indentation d'un texte
     * @param  string $text  Le texte à traiter
     * @param  Integer $nb   Le nombre d'indentation à appliquer
     * @return string        Le rendu avec indentation
     */
    private function renderIndent($text, $nb): string
    {
        $indent = '    ';
        $render = '';

        $real_indent = '';

        for ($i = 1; $i <= $nb; $i++) {
            $real_indent .= $indent;
        }

        $lines = explode(PHP_EOL, $text);

        foreach ($lines as $line) {

            if (!empty($line)) {
                $render .= $real_indent . $line . PHP_EOL;
            }
        }

        return $render;
    }



    /**
     * Mise à jour des attributs depuis le tableau passé en paramètre
     * @param Array $attrs Liste des attributs à traiter
     * @return void
     */
    private function setAttrs($attrs)
    {
        $this->resetAttrs();
        if (is_null($attrs)) {
            return ;
        }

        // Liste des attributs autorisés
        $attrs_available = array_map(function ($val) {
            return trim($val, '_');
        }, array_keys(get_object_vars($this)));

        // Traitement des attributs
        if (is_array($attrs)) {
            foreach ($attrs as $attr => $value) {
                if (in_array($attr, $attrs_available)) {
                    if (is_array($value)) {
                        $this->{'_' . $attr} = $value;
                    } else {
                        $this->{$attr} = $value;
                    }
                } else {
                    throw new \InvalidArgumentException("Le paramètre [" . $attr . "] n'est pas pris en charge");
                }
            }
        } else {
            throw new \InvalidArgumentException("Type de paramètre invalide. Attendu : Tableau");
        }

    }

    /**
     * Réinitialisation des attributs de constructions
     * @return void
     */
    private function resetAttrs()
    {
        $attrs = array_keys(get_object_vars($this));

        foreach ($attrs as $attr) {

            if (preg_match('/^_.*/', $attr)) {
                $this->{$attr} = [];
            } else if (preg_match('/^is_.*/', $attr)) {
                $this->{$attr} = false;
            } else {
                $this->{$attr} = null;
            }
        }
    }
}
